import unittest
from typing import List

from argument import Argument
from arguments import Arguments
from expression import Expression
from expression_to_values import maximally_resolve_expression
from file_path import FilePath
from file_range import parse
from literal import Literal
from name import Name
from parameter import Parameter
from parameter_declaration import ParameterDeclaration
from parameter_declarations import ParameterDeclarations
from subprogram_declaration import SubprogramDeclaration


class TestExpressionToValues(unittest.TestCase):
    def test_closed_expression_is_singleton_value(self):
        file_path = FilePath("a file path")
        subprogram_declaration = SubprogramDeclaration(Name("function name"),
                                                       file_path,
                                                       parse("1:1-2:2"),
                                                       ParameterDeclarations())
        expression = Literal("literal")
        assert expression.is_closed()
        expected = [expression]
        self.assertEqual(expected, maximally_resolve_expression(expression, subprogram_declaration))

    def test_open_expression_in_uncalled_function_is_that_expression(self):
        name: Name = Name("name")
        file_path = FilePath("a file path")
        parameter = ParameterDeclaration(name, file_path, parse("2:2-3:3"), Name("a type"))
        parameters = ParameterDeclarations()
        parameters.add(parameter)
        subprogram_declaration = SubprogramDeclaration(Name("function name"),
                                                       file_path,
                                                       parse("1:1-10:10"),
                                                       parameters)
        expression = Parameter(Name("parameter_declaration name"), file_path, parse("4:4-5:5"))
        assert not expression.is_closed()
        expected = [expression]
        self.assertEqual(expected, maximally_resolve_expression(expression, subprogram_declaration))

    def test_open_expression_closed_by_single_function_call(self):
        expression: Expression
        value: Expression
        subprogram_declaration: SubprogramDeclaration
        actual: List[Expression]

        def setup() -> None:
            nonlocal expression, value, subprogram_declaration

            name: Name = Name("name")
            file_path = FilePath("a file path")
            a_file_range = parse("2:2-3:3")
            parameter = ParameterDeclaration(name, file_path, a_file_range, Name("a type"))
            expression = parameter.to_parameter_reference()
            parameters = ParameterDeclarations()
            parameters.add(parameter)
            subprogram_declaration = SubprogramDeclaration(Name("function name"),
                                                           file_path,
                                                           a_file_range,
                                                           parameters)
            value = Literal("value")
            assert value.is_closed()
            arguments = Arguments()
            arguments.add(Argument(name, value))
            subprogram_declaration.add_call(subprogram_declaration,
                                            arguments,
                                            a_file_range)

        def exercise() -> None:
            nonlocal actual
            actual = maximally_resolve_expression(expression, subprogram_declaration)

        def verify() -> None:
            expected = [value]
            self.assertEqual(expected, actual)

        setup()
        exercise()
        verify()

    def test_open_expression_with_single_open_function_call_has_no_value(self):
        expression: Expression
        substitution: Expression
        subprogram_declaration: SubprogramDeclaration
        actual: List[Expression]

        def setup() -> None:
            nonlocal expression, substitution, subprogram_declaration

            name: Name = Name("name")
            file_path = FilePath("a file path")
            a_file_range = parse("2:2-3:3")
            parameter_declaration = ParameterDeclaration(name, file_path, a_file_range, Name("a type"))
            expression = parameter_declaration.to_parameter_reference()
            assert not expression.is_closed()
            parameters = ParameterDeclarations()
            parameters.add(parameter_declaration)
            subprogram_declaration = SubprogramDeclaration(Name("function name"),
                                                           file_path,
                                                           a_file_range,
                                                           parameters)
            substitution = Parameter(Name("parameter_declaration name"), file_path, a_file_range)
            assert not substitution.is_closed()
            arguments = Arguments()
            arguments.add(Argument(name, substitution))
            subprogram_declaration.add_call(subprogram_declaration,
                                            arguments,
                                            a_file_range)

        def exercise() -> None:
            nonlocal actual
            actual = maximally_resolve_expression(expression, subprogram_declaration)

        def verify() -> None:
            self.assertEqual([substitution], actual)

        setup()
        exercise()
        verify()

    def test_open_expression_closed_by_multiple_direct_function_calls(self):
        expression: Expression
        values: List[Expression]
        subprogram_declaration: SubprogramDeclaration
        actual: List[Expression]

        def setup() -> None:
            nonlocal expression, values, subprogram_declaration

            name: Name = Name("name")
            file_path = FilePath("a file path")
            a_file_range = parse("2:2-3:3")
            parameter = ParameterDeclaration(name, file_path, a_file_range, Name("a type"))
            expression = parameter.to_parameter_reference()
            parameters = ParameterDeclarations()
            parameters.add(parameter)
            subprogram_declaration = SubprogramDeclaration(Name("function name"),
                                                           file_path,
                                                           a_file_range,
                                                           parameters)
            values = [Literal("value"),
                      Literal("another value"),
                      Literal("yet another")]
            for value in values:
                assert value.is_closed()
                arguments = Arguments()
                arguments.add(Argument(name, value))
                subprogram_declaration.add_call(subprogram_declaration,
                                                arguments,
                                                a_file_range)

        def exercise() -> None:
            nonlocal actual

            actual = maximally_resolve_expression(expression, subprogram_declaration)

        def verify() -> None:
            self.assertEqual(values, actual)

        setup()
        exercise()
        verify()

    def test_open_expression_closed_by_single_chain_of_function_calls(self):
        expression: Expression
        value: Expression
        subprogram_declaration: SubprogramDeclaration
        actual: List[Expression]

        def setup() -> None:
            nonlocal expression, value, subprogram_declaration

            # subprogram
            name: Name = Name("name")
            file_path = FilePath("a file path")
            a_file_range = parse("2:2-3:3")
            parameter = ParameterDeclaration(name, file_path, a_file_range, Name("a type"))
            expression = parameter.to_parameter_reference()
            parameters = ParameterDeclarations()
            parameters.add(parameter)
            subprogram_declaration = SubprogramDeclaration(Name("function name"),
                                                           file_path,
                                                           a_file_range,
                                                           parameters)

            # another subprogram
            another_name: Name = Name("another name")
            another_file_path = FilePath("another file path")
            another_parameter = ParameterDeclaration(another_name, another_file_path, a_file_range, Name("a type"))
            another_parameters = ParameterDeclarations()
            another_parameters.add(another_parameter)

            another_subprogram_declaration = SubprogramDeclaration(Name("another function name"),
                                                                   another_file_path,
                                                                   a_file_range,
                                                                   another_parameters)

            # calls
            open_arguments = Arguments()
            open_arguments.add(Argument(name, another_parameter.to_parameter_reference()))
            another_subprogram_declaration.add_call(subprogram_declaration,
                                                    open_arguments,
                                                    a_file_range)

            value = Literal("value")
            assert value.is_closed()
            closed_arguments = Arguments()
            closed_arguments.add(Argument(another_name, value))
            another_subprogram_declaration.add_call(another_subprogram_declaration,
                                                    closed_arguments,
                                                    a_file_range)

        def exercise() -> None:
            nonlocal actual
            actual = maximally_resolve_expression(expression, subprogram_declaration)

        def verify() -> None:
            self.assertEqual([value], actual)

        setup()
        exercise()
        verify()


if __name__ == '__main__':
    unittest.main()
